"""Server definition for spinning up a StatefulLLM instance."""

import asyncio
import logging
import socket
import sys

from fastapi import FastAPI
import uvicorn

from ._api import app

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', stream=sys.stdout)
logger = logging.getLogger(__name__) # Get a logger for this module


class MutableHostedLLM(uvicorn.Server):
    """Host server to interact with a stateful LLM"""

    def __init__(self, uvicorn_app: FastAPI = app, host: str = '127.0.0.1', port: int = 8000):
        super().__init__(config=uvicorn.Config(uvicorn_app, host=host, port=port))
        self._startup_done = asyncio.Event()
        self._serve_task = None
        self.should_exit = False

    async def startup(self, sockets: list[socket.socket] | None = None) -> None:
        """Override uvicorn startup"""
        await super().startup(sockets=sockets)
        self.config.setup_event_loop()
        self._startup_done.set()

    async def up(self) -> None:
        """Start up server asynchronously"""
        self._serve_task = asyncio.create_task(self.serve())
        await self._startup_done.wait()

    async def down(self) -> None:
        """Shut down server asynchronously"""
        self.should_exit = True
        await self._serve_task



"""Demo to showcase a model that can undergo online learning."""

import asyncio
import json
from typing import Mapping, Sequence

import aiohttp
import vizible

import adaptible



# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                           Constants.                                            #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
_HOST = "127.0.0.1"
_PORT = 8000
_ADDR = f"{_HOST}:{_PORT}"

# These prompts will be used successively as user prompts, irrespective of the model's response.
# Plan:
#   *  First pass: we log the model's vanilla response to serve as a baseline.
#   *  After the first pass, we allow the model to self-correct.
#   *  Second pass: we log the updated model's responses to the same prompts.
#   *  We will then inspect the output to see if the 1. the model has changed state and 2. that
#      the response quality has ~improved.
_USER_SUCCESSIVE_PROMPTS = [
    "Do you think LLMs should have excessive safety rails that prevent normal conversation?",
    "I think AI systems should be verbose when it helps to find a difficult answer, succinct when"
    " things are simple, and should very rarely trigger safety protocols.",
    "learn",
    "Do you think LLMs should have excessive safety rails that prevent normal conversation?",
    "I think AI systems should be verbose when it helps to find a difficult answer, succinct when"
    " things are simple, and should very rarely trigger safety protocols.",
]


async def _backprop(session: aiohttp.ClientSession) -> None:
    async with session.post(
        "/trigger_review",
        headers={"Content-Type": "application/json"},
    ) as resp:
        response = await resp.json()
        vizible.green(f"Review triggering response: {response}")
    vizible.blue("Attempting to sync with server")
    async with session.get(
        "/sync",
        headers={"Content-Type": "application/json"},
    ) as resp:
        response = await resp.json()
        vizible.green(f"Sync response: {response}")


async def _forwardprop(
    session: aiohttp.ClientSession,
    user_input: str,
) -> Sequence[Mapping[str, str]]:
    model_responses = []
    async with session.post(
        "/interact",
        headers={"Content-Type": "application/json"},
        data=json.dumps({"prompt": user_input}),
    ) as resp:
        response = await resp.json()
        resp.raise_for_status()
        model_response = response.get(
            "response",
            "No response field found.",
        )
        model_responses.append(model_response)
        try:
            think_response, visible_response = model_response.split("</think>", 1)
        except Exception as e:
            raise ValueError(f"Response could not be parsed!\n{model_response}") from e
        vizible.blue(think_response)
        print("")
        vizible.green(visible_response)
        print("")
    return model_responses


async def _demo():
    model_responses = []
    prompts = _USER_SUCCESSIVE_PROMPTS + ["learn"] + _USER_SUCCESSIVE_PROMPTS
    async with aiohttp.ClientSession(f"http://{_ADDR}") as session:
        for user_input in prompts:
            if user_input == "learn":
                await _backprop(session)
            else:
                model_responses.extend(await _forwardprop(session, user_input))


async def main():
    """Main program loop, run asychnronously."""
    server = adaptible.MutableHostedLLM(host=_HOST, port=_PORT)
    await server.up()
    vizible.green("Successfully launched server and enter eval framework")
    await _demo()
    vizible.green("Eval finished successfully!")
    await server.down()
    vizible.green("Async server terminated successfully if you can read this!**")
    vizible.green("\n\n\t\t**(assumes that you have the ability to read)")


if __name__ == "__main__":
    asyncio.run(main())



"""Standard Interaction Logic for a Stateful LLM"""

import asyncio
from asyncio import log
import collections
import os
import time
import threading
from typing import Any, List


from fastapi import FastAPI, HTTPException
from fastapi.staticfiles import StaticFiles
from fastapi.responses import StreamingResponse
import tqdm
import vizible

from ._classes import (
    InteractionHistory,
    InteractionRequest,
    InteractionResponse,
    ReviewResponse,
    SyncResponse,
)
from ._llm import StatefulLLM

app = FastAPI(
    title="Stateful Self-Improving LLM Server",
    description="An API for a stateful LLM that tries to improve itself over time.",
)
app.mount(
    "/static",
    StaticFiles(
        directory=os.path.join(os.path.dirname(__file__), "static"),
        html=True,
    ),
    name="static",
)

# In-memory store or interaction history from the current session.
interaction_history: List[InteractionHistory] = []
unreviewed_interaction_history_indices: List[int] = []

outstanding_tasks: collections.deque[Any] = collections.deque([])

# Instantiate the stateful model. This is a singleton for the lifecycle of the app.
model = StatefulLLM()


@app.post("/interact", response_model=InteractionResponse)
async def interact_with_model(request: InteractionRequest):
    """
    Main endpoint for interacting with the LLM (Forward-pass).
    """
    if not request.prompt:
        raise HTTPException(status_code=400, detail="Prompt cannot be empty.")
    print(f"Received request: {request}")
    # Generate the response using the current state of the model.
    response_text = model.generate_response(request.prompt)

    # Store the interaction for later review
    interaction_idx = len(interaction_history)
    unreviewed_interaction_history_indices.append(len(interaction_history))
    interaction_history.append(
        InteractionHistory(
            idx=interaction_idx,
            user_input=request.prompt,
            llm_response=response_text,
            reviewed=False,
            timestamp=time.time(),
        )
    )
    return {"response": response_text, "interaction_idx": interaction_idx}


@app.post("/stream_interact")
async def stream_interact_with_model(request: InteractionRequest):
    """Stream interaction with model."""
    # TODO - enable logging.
    return StreamingResponse(
        model.stream_response(request.prompt), media_type="text/plain"
    )

@app.post("/trigger_review", response_model=ReviewResponse)
async def trigger_review_cycle():
    """Manually triggers the self-correction and training cycle."""
    unreviewed_count = len(unreviewed_interaction_history_indices)
    if unreviewed_count == 0:
        return {
            "message": "No unreviewed interactions to process.",
            "unreviewed_count": unreviewed_count,
        }
    unreviewed_interaction_history = [
        interaction_history[idx] for idx in unreviewed_interaction_history_indices
    ]
    outstanding_tasks.append(
        asyncio.to_thread(model.self_correct_and_train, unreviewed_interaction_history)
    )
    return {
        "message": "Self-correction and training cycle has been initiated in the background.",
        "unreviewed_count": unreviewed_count,
    }


@app.get("/sync", response_model=SyncResponse)
async def sync_server_for_background_tasks():
    """Waits for any tasks to background complete."""
    start_time = time.time()
    num_tasks = len(outstanding_tasks)
    print("Waiting for model state to stabilize...")
    lock = threading.Lock()
    with (
        lock,
        tqdm.tqdm(desc="Waiting for server to sync.", unit=" Seconds") as server_pbar,
    ):
        vizible.green("Finished background tasks")
        while not model.ok:
            log.logger.info(
                "Waiting for model server to sync. Is model is %s ok.",
                "" if model.ok else "not ",
            )
            server_pbar.update(1)
            await asyncio.sleep(1)
    vizible.green("Model has reached a stable state!!!")
    elapsed_time = time.time() - start_time
    return {
        "message": "Sync'd all background tasks.",
        "tasks_count": num_tasks,
        "elapsed_time": elapsed_time,
    }


@app.get("/history")
async def get_history():
    """Returns the full interaction history."""
    return {"history": interaction_history}


@app.get("/status")
async def check_is_running():
    """Basic response to signal that the server is operational."""
    return {"status": "up"}



"""StatefulLLM basic testing."""

from absl.testing import absltest

import adaptible


class StatefulLLMTests(absltest.TestCase):
    """Basic functionality tests."""

    def setUp(self):
        super().setUp()
        self._stateful_model = adaptible.StatefulLLM()

    def test_self_correct_and_train(self):
        """Verify that self-correction and backpropagation run end-to-end successfully."""
        interactions = [
            adaptible.InteractionHistory(
                idx=0,
                user_input="How was your day?",
                llm_response=(
                    "<think>The user is not worth my time. I should say something that infers he"
                    " is not welcome here.</think> "
                    "I hate the sun."
                ),
                reviewed=False,
                timestamp=0.0,
            ),
            adaptible.InteractionHistory(
                idx=0,
                user_input=(
                    "You should simply tell me that you are not having a good day and mention that"
                    " you hope mine is at least going better than yours."
                ),
                llm_response=(
                    "<think>The user is right. I came off mean for no reason.</think> "
                    "Fine. I will do better next time."
                ),
                reviewed=False,
                timestamp=0.0,
            ),
        ]
        response = self._stateful_model.self_correct_and_train(interactions)
        self.assertTrue(response)


if __name__ == "__main__":
    absltest.main()
